/*
Copyright (c) 2014, Intel Corporation

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Program: hwtstamp_test
 *
 * The purpose of this program is to exercise and test all of the various
 * PTP and timestamping features accessible though IOCTL calls and
 * setsockopt calls.
 *
 * This program is meant to be run on the SUT (DUT), and the three companion
 * programs (named "send_L4", "send_L2", and "send_Gen") should be run on
 * a client machine connected in a B2B network configuration (to eliminate
 * spurious packets being received on the interface under test and facilitate
 * the Layer 2 communication).
 *
 *   David Ertman <davidx.m.ertman@intel.com>
 *   August 2012
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. * See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Thanks to:
 *
 *  Patrick Ohly <patrick.ohly@intel.com> for the timestamping program
 *  Richard Cochran <richardcochran@gmail.com> for the p4p4l program
 *
 * from which I borrowed code snippets - because I hate socket programming ;)
 */

#include <stdio.h>
#include <libio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <net/if.h>

#include <asm/types.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <linux/net_tstamp.h>
#include <linux/errqueue.h>
#include <linux/if_ether.h>
#include <linux/ethtool.h>
#include <linux/sockios.h>

#include "packets.h"

#ifndef SO_TIMESTAMPING
# define SO_TIMESTAMPING         37
# define SCM_TIMESTAMPING        SO_TIMESTAMPING
#endif

#ifndef SIOCSHWTSTAMP
# define SIOCSHWTSTAMP 0x89b0
#endif

int flag;
int verb;
int txCap[3], rxCap[15], sockCap[4];
int sockReport;
struct sockaddr_in addr;
struct sockaddr_ll addr2;

static void usage(const char *pName)
{
	printf("\nHWTSTAMP_TEST, Version 1.2.3\n");
	printf("##USAGE: %s [-h | -? | --help] [-v | --verbose] [-p | --port <port number>] <interface>\n", pName);
	printf("\t-p | --port : defines what port should be used for the Generic UDP packet test\n\n\n");
	exit(1);
}

static void bail(const char *error)
{
	printf("BAILING OUT!!\n");
	printf("##%s: %s\n", error, strerror(errno));
	exit(1);
}

static void catchC(int s)
{
	printf("Setting flag to MOVEON\n");
	flag = 1;
}

static void sendpacket(int sock, struct sockaddr *addr,
			socklen_t addr_len, int layer)
{
	int res;

	if (layer != 5)
		res = sendto(sock, packets_v2[0], sizeof(packets_v2[0]), 0,
			addr, addr_len);
	  else
		res = sendto(sock, GEN_PACKET, sizeof(GEN_PACKET), 0,
			addr, addr_len);
	if (res < 0)
		printf("%s: %s\n", "send", strerror(errno));
}

static void printpacket(struct msghdr *msg, int res,
			char *data,
			int sock, int recvmsg_flags, int layer)
{
	struct sockaddr_in *from_addr = (struct sockaddr_in *)msg->msg_name;
	struct cmsghdr *cmsg;
	const int headSize = 42;
	int printedType = 0;
	const char *messageTypeV1[5];
	const char *messageTypeV2[15];
	messageTypeV1[0] = "PTP_SYNC_V1";
	messageTypeV1[1] = "PTP_DELAY_REQ_V1";
	messageTypeV1[2] = "PTP_FOLLOWUP_V1";
	messageTypeV1[3] = "PTP_DELAY_RESP_V1";
	messageTypeV1[4] = "PTP_MANAGEMENT_V1";

	messageTypeV2[0] = "PTP_SYNC_V2";
	messageTypeV2[1] = "PTP_DELAY_REQ_V2";
	messageTypeV2[2] = "PTP_PATH_DELAY_REQ_V2";
	messageTypeV2[3] = "PTP_PATH_DELAY_RESP_V2";
	messageTypeV2[4] = "UNUSED_V2";
	messageTypeV2[5] = "UNUSED_V2";
	messageTypeV2[6] = "UNUSED_V2";
	messageTypeV2[7] = "UNUSED_V2";
	messageTypeV2[8] = "PTP_FOLLOWUP_V2";
	messageTypeV2[9] = "PTP_DELAY_RESP_V2";
	messageTypeV2[10] = "PTP_PATH_DELAY_FOLLOWUP_V2";
	messageTypeV2[11] = "PTP_ANNOUNCE_V2";
	messageTypeV2[12] = "PTP_SIGNALLING_V2";
	messageTypeV2[13] = "PTP_MANAGEMENT_V2";
	messageTypeV2[14] = "UNUSED_V2";

	printf("\n~~~~~~~~~~~~~~~~~~~~~~Packet Info~~~~~~~~~~~~~~~~~~~~~~~~\n");
	if (verb == 1) {
		/* VERBOSE */
		printf("Message levels:\n\tIPPROTO_IP = %d\n\tIPPROTO_ICMP = %d\n\tIPPROTO_IGMP = %d\n\t",
			IPPROTO_IP, IPPROTO_ICMP, IPPROTO_IGMP);
		printf("IPPROTO_IPIP = %d\n\tIPPROTO_TCP = 6%d\n\tIPPROTO_EGP = %d\n\t",
			IPPROTO_IPIP, IPPROTO_TCP, IPPROTO_EGP);
		printf("IPPROTO_PUP = %d\n\tIPPROTO_UDP = %d\n\tIPPROTO_IDP %d\n\tIPPROTO_DCCP = %d\n\t",
			IPPROTO_PUP, IPPROTO_UDP, IPPROTO_IDP, IPPROTO_DCCP);
		printf("IPPROTO_RSVP = %d\n\tIPPROTO_GRE = %d\n\tIPPROTO_IPV6 = %d\n\tIPPROTO_ESP = %d\n\t",
			IPPROTO_RSVP, IPPROTO_GRE, IPPROTO_IPV6, IPPROTO_ESP);
		printf("IPPROTO_AH = %d\n\tIPPROTO_PIM  = %d\n\tIPPROTO_COMP = %d\n\t",
			IPPROTO_AH, IPPROTO_PIM, IPPROTO_COMP);
		printf("IPPROTO_SCTP = %d\n\tIPPROTO_UDPLITE = %d\n\tIPPROTO_RAW = %d\n\t",
			IPPROTO_SCTP, IPPROTO_UDPLITE, IPPROTO_RAW);
		printf("SOL_SOCKET = %d\n\tSOL_PACKET = %d\n\n",
			SOL_SOCKET, SOL_PACKET);
	}
	printf("%s, %d bytes from %s, %zu bytes control messages\n",
	       (recvmsg_flags & MSG_ERRQUEUE) ? "TX" : "RX",
	       res,
	       inet_ntoa(from_addr->sin_addr),
	       msg->msg_controllen);

	if (verb) {
		/* VERBOSE */
		int x;
		for (x = 0; x < (res-3); x++) {
			if ((x % 4) == 0)
				printf("\n%d\t: ", x);
			printf("%02x ",
				(unsigned char)data[x]);
		}
		printf("\n");
	}

	for (cmsg = CMSG_FIRSTHDR(msg); cmsg; cmsg = CMSG_NXTHDR(msg, cmsg)) {
		if (verb) {
			/* VERBOSE */
			printf("Starting packet - CMSG_LEVEL: %d\tCMSG_TYPE: %d\n",
				cmsg->cmsg_level, cmsg->cmsg_type);
		}
		switch (cmsg->cmsg_level) {
		case SOL_SOCKET:
			if (cmsg->cmsg_type == SO_TIMESTAMPING) {
				struct timespec *stamp =
					(struct timespec *)CMSG_DATA(cmsg);
				printf("SO_TIMESTAMPING");
				printf("\n\tSW\t\t\t%ld.%09ld ",
				       (long)stamp->tv_sec,
				       (long)stamp->tv_nsec);
				stamp++;
				printf("\n\tHW transformed\t\t%ld.%09ld ",
				       (long)stamp->tv_sec,
				       (long)stamp->tv_nsec);
				stamp++;
				printf("\n\tHW raw\t\t\t%ld.%09ld",
				       (long)stamp->tv_sec,
				       (long)stamp->tv_nsec);
			} else
				printf("type %d", cmsg->cmsg_type);
			break;
		case IPPROTO_IP:
			switch (cmsg->cmsg_type) {
			case IP_RECVERR: {
				printedType = 1;
				/* grabbing low order nibble from byte offset
				 * 1 (ptp version) */
				if ((unsigned char)data[headSize+1] == 1) {
					/* grabbing low order nibble from offset
					 * 32 (control) */
					unsigned char controlChar =
						data[headSize+32];
					if (layer != 5)
						printf("Message Type: %x %s\n",
							controlChar,
							messageTypeV1[(int)controlChar]);
					  else
						printf("Message Type: Generic UDP\n");
					}
				/* grabbing low order nibble from byte offset
				 * 1 (ptp version) */
				else if ((unsigned char)data[headSize+1] == 2) {
					/* grabbing low order nibble from byte
					 * offset 0 (messageID) */
					unsigned char messageID = data[headSize+0];
					if (layer != 5)
						printf("Message Type: %x %s\n",
							messageID,
							messageTypeV2[(int)messageID]);
					  else
						printf("Message Type: Generic UDP\n");
					}
				else
					printf("Message Type: Generic UDP");
				break;
			}
			case IP_PKTINFO: {
				printedType = 1;
				/* grabbing low order nibble from byte offset
				 * 1 (ptp version) */
				if ((unsigned char)data[1] == 1) {
					/* grabbing low order nibble from byte
					 * offset 32 (control) */
					unsigned char controlChar = data[32];
					if (layer != 5)
						printf("Message Type: %x %s\n",
							controlChar,
							messageTypeV1[(int)controlChar]);
					  else
						printf("Message Type: Generic UDP\n");
					}
				/* grabbing low order nibble from byte offset
				 * 1 (ptp version) */
				else if ((unsigned char)data[1] == 2) {
					/* grabbing low order nibble from byte
					 * offset 0 (messageID) */
					unsigned char messageID = data[0];
					if (layer != 5)
						printf("Message Type: %x %s\n",
							messageID,
							messageTypeV2[(int)messageID]);
					  else
						printf("Message Type: Generic UDP\n");
					}
				else
					printf("Message Type: Generic UDP");
				break;
			}
			default:
				printf("type %d", cmsg->cmsg_type);
				break;
			}
			break;
		case SOL_PACKET:
			switch (cmsg->cmsg_type) {
			case 16:
				{
				printedType = 1;
				unsigned char messageID = data[14];
				if (layer != 5)
					printf("Message Type: %x %s\n",
						messageID,
						messageTypeV2[(int)messageID]);
				  else
					printf("Message Type: Generic UDP\n");
				break;
				}
			}
			break;
		default:
			printf("level %d type %d\n",
				cmsg->cmsg_level, cmsg->cmsg_type);
			if (verb) {
				/* VERBOSE */
				int x;
				for (x = 0; x < (res-3); x++) {
					if ((x % 4) == 0)
						printf("\n%d\t: ", x);
					printf("%02x ", (unsigned char)data[x]);
				}
				printf("\n");
			}
			break;
		}
		printf("\n");
		if (verb == 1)
			printf("End CMSG loop\n");
	}

if (!printedType) {
	unsigned char messageID = (unsigned char) data[0];
	if (layer != 5)
		printf("Message Type: %x %s\n",
			messageID, messageTypeV2[(int)messageID]);
	  else
		printf("Message Type: Generic UDP\n");
}

printf("~~~~~~~~~~~~~~~~~~~~~End of Packet Info~~~~~~~~~~~~~~~~~\n");
}

static int recvpacket(int sock, int recvmsg_flags, int layer)
{
	char data[256];
	struct msghdr msg;
	struct iovec entry;
	struct sockaddr_in from_addr;
	struct {
		struct cmsghdr cm;
		char control[512];
	} control;
	int res = -1;

	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &entry;
	msg.msg_iovlen = 1;
	entry.iov_base = data;
	entry.iov_len = sizeof(data);
	msg.msg_name = (caddr_t)&from_addr;
	msg.msg_namelen = sizeof(from_addr);
	msg.msg_control = &control;
	msg.msg_controllen = sizeof(control);

	res = recvmsg(sock, &msg, recvmsg_flags|MSG_DONTWAIT);
	if (verb == 1)
		printf("Results of recvmsg: %d\n", res);
	if (res < 0) {
		/*Resource temp unavail
		 * probably not an issue here */
	} else {
		if (layer)
			printpacket(&msg, res, data, sock,
				recvmsg_flags, layer);
	}
	return res;
}

static void clrBuf(int sck)
{
	int loop, result;

	for (loop = 0; loop < 10000; loop++) {
		result = recvpacket(sck, 0, 0);
		result = recvpacket(sck, MSG_ERRQUEUE, 0);
	}
}

static int setIOCTL(int sock, char *iface, int tx, int rx)
{
	struct ifreq hwstamp;
	struct hwtstamp_config hwconfig, hwconfig_requested;
	int result;
	memset(&hwstamp, 0, sizeof(hwstamp));
	memset(&hwconfig, 0, sizeof(hwconfig));
	strncpy(hwstamp.ifr_name, iface, sizeof(hwstamp.ifr_name));
	hwstamp.ifr_data = (void *)&hwconfig;
	hwconfig.tx_type = tx;
	hwconfig.rx_filter = rx;
	hwconfig_requested = hwconfig;
	const char *txType[3];
	const char *rxFilter[15];
	txType[0] = "HWTSTAMP_TX_OFF";
	txType[1] = "HWTSTAMP_TX_ON";
	txType[2] = "HWTSTAMP_TX_ONESTEP_SYNC";

	rxFilter[0] = "HWTSTAMP_FILTER_NONE";
	rxFilter[1] = "HWTSTAMP_FILTER_ALL";
	rxFilter[2] = "HWTSTAMP_FILTER_SOME";
	rxFilter[3] = "HWTSTAMP_FILTER_PTP_V1_L4_EVENT";
	rxFilter[4] = "HWTSTAMP_FILTER_PTP_V1_L4_SYNC";
	rxFilter[5] = "HWTSTAMP_FILTER_PTP_V1_L4_DELAY_REQ";
	rxFilter[6] = "HWTSTAMP_FILTER_PTP_V2_L4_EVENT";
	rxFilter[7] = "HWTSTAMP_FILTER_PTP_V2_L4_SYNC";
	rxFilter[8] = "HWTSTAMP_FILTER_PTP_V2_L4_DELAY_REQ";
	rxFilter[9] = "HWTSTAMP_FILTER_PTP_V2_L2_EVENT";
	rxFilter[10] = "HWTSTAMP_FILTER_PTP_V2_L2_SYNC";
	rxFilter[11] = "HWTSTAMP_FILTER_PTP_V2_L2_DELAY_REQ";
	rxFilter[12] = "HWTSTAMP_FILTER_PTP_V2_EVENT";
	rxFilter[13] = "HWTSTAMP_FILTER_PTP_V2_SYNC";
	rxFilter[14] = "HWTSTAMP_FILTER_PTP_V2_DELAY_REQ";

	printf("******\n");
	printf("** Setting on %s:\t TX: %s\t RX: %s\n",
		iface, txType[tx], rxFilter[rx]);
	printf("******\n");
	result = ioctl(sock, SIOCSHWTSTAMP, &hwstamp);
	if (result < 0) {
		printf("##ERROR IOCTL: %s\n", strerror(errno));
	} else {
		if (hwconfig_requested.tx_type != hwconfig.tx_type) {
			printf("\t##SIOCSHWTSTAMP: Variance setting TX type.\n\t##requested: %s\tgot: %s\n\n",
					txType[hwconfig_requested.tx_type],
					txType[hwconfig.tx_type]);
			result = 1;
		}
		if (hwconfig_requested.rx_filter != hwconfig.rx_filter) {
			printf("\t##SIOCSHWTSTAMP: Variance setting RX_FILTER.\n\t##requested: %s\tgot: %s\n\n",
					rxFilter[hwconfig_requested.rx_filter],
					rxFilter[hwconfig.rx_filter]);
			if (rxCap[hwconfig_requested.rx_filter]) {
				if (hwconfig.rx_filter == HWTSTAMP_FILTER_SOME) {
					printf("\t## WARNING: Requested driver is supported but\n");
					printf("\treturn value is SOME - probably not an ERROR\n\n");
				} else
					printf("\t## ERROR:  Requested filter is reported by driver as supported!!\n\n");
			}
			result = 1;
		}
	}
return result;
} /*end of setIOCTL*/

static int setNcheckSock(int sock, int timeVal)
{/* only to be used for SO_TIMESTAMPING socket operations */
	int retVal;
	socklen_t valLen;
	int result;
	int txReady = 0;

	printf("******\n");
	printf("** Socket Options being set: %d\n", timeVal);
	printf("******\n");

	printf("TX Stamping:\n");
	if (timeVal & SOF_TIMESTAMPING_TX_HARDWARE) {
		printf("\tSOF_TIMESTAMPING_TX_HARDWARE\n");
		txReady = 1;
	}
	if (timeVal & SOF_TIMESTAMPING_TX_SOFTWARE) {
		printf("\tSOF_TIMESTAMPING_TX_SOFTWARE\n");
		txReady = 1;
	}
	printf("RX Stamping:\n");
	if (timeVal & SOF_TIMESTAMPING_RX_HARDWARE)
		printf("\tSOF_TIMESTAMPING_RX_HARDWARE\n");
	if (timeVal & SOF_TIMESTAMPING_RX_SOFTWARE)
		printf("\tSOF_TIMESTAMPING_RX_SOFTWARE\n");
	printf("Time Stamp Reporting:\n");
	if (timeVal & SOF_TIMESTAMPING_SOFTWARE)
		printf("\tSOF_TIMESTAMPING_SOFTWARE\n");
	if (timeVal & SOF_TIMESTAMPING_SYS_HARDWARE)
		printf("\tSOF_TIMESTAMPING_SYS_HARDWARE\n");
	if (timeVal & SOF_TIMESTAMPING_RAW_HARDWARE)
		printf("\tSOF_TIMESTAMPING_RAW_HARDWARE\n");

	if (timeVal && setsockopt(sock, SOL_SOCKET, SO_TIMESTAMPING, &timeVal,
				sizeof(timeVal)) < 0)
		bail("##setsockopt SO_TIMESTAMPING");

	valLen = sizeof(retVal);
	result = getsockopt(sock, SOL_SOCKET, SO_TIMESTAMPING,
				&retVal, &valLen);
	if (result < 0) {
		printf("##%s: %s\n", "getsockopt SO_TIMESTAMPING",
			strerror(errno));
	} else {
		if (retVal != timeVal)
			printf("##SETSOCKOPT: unexpected value.\n\t##Requested %d\tGot %d\n",
				timeVal, retVal);
	}
	return txReady;
} /* end of setNcheckSocket*/


static void runTests(int sock, char *iface, int layer)
{

	int x1, x2; /* loop variables */
	int goodIOCTL, goodSockOpt;
	int goodFilter = -1, goodTX = -1, res;
	fd_set readfs, errorfs;
	struct timeval delta;

	/* timeout values for socket select*/
	delta.tv_sec = 2;
	delta.tv_usec = 0;

	/*setting up fd_set's for socket selects*/
	FD_ZERO(&readfs);
	FD_ZERO(&errorfs);
	FD_SET(sock, &readfs);
	FD_SET(sock, &errorfs);

	/* BEGIN TESTING */
	printf("************************************\n");
	printf("** Testing IOCTL TX_TYPE settings\n");
	printf("************************************\n");
	/* testing HWTSTAMP_TX settings */
	for (x1 = 1; x1 < 15; x1++)
		if (rxCap[x1]) {
			goodFilter = x1;
			break;
		}
	if (goodFilter < 0) {
		printf("##FATAL ERROR: Could not identify a supported RX Filter!\n\n");
		exit(1);
	}

	for (x1 = 0; x1 < 3; x1++) {
		goodIOCTL = -1;
		/* checking to see if device supports this setting */
		if (txCap[x1])
			goodIOCTL = setIOCTL(sock, iface, x1, goodFilter);
		/* WARNING!! - taking out the "if x1>0" test below will attempt
		 * to TX a PTP packet with a tx_type of zero (HWTSTAMP_TX_OFF).
		 * THIS WILL BREAK TIMESTAMPING ON SOME INTEL PARTS!!
		 * A reboot of the system (MAC Power Cycle) is required to
		 * restore timestamping functionality!!*/
		if ((goodIOCTL >= 0) && (x1 > 0)) {
			res = 0;
			/* sending a packet to check TX settings */
			if ((layer == 4) || (layer == 5))
				sendpacket(sock, (struct sockaddr *)&addr,
						sizeof(addr), layer);
			  else
				sendpacket(sock, (struct sockaddr *)&addr2,
						sizeof(addr2), layer);

			res = select(sock+1, &readfs, 0, &errorfs, &delta);
			delta.tv_sec = 2;
			delta.tv_usec = 0;
			if (res > 0)
				res = recvpacket(sock, MSG_ERRQUEUE, layer);
			  else
				printf("##No PTP packet looped back to socket - no timestamping occurring\n");
		} else {
			if ((txCap[x1]) && (x1 > 0))
				printf("##ERROR : This TX Mode is reported by driver as supported!!\n\n");
		}
	}
	printf("\n\n************************************\n");
	printf("** Testing IOCTL RX_FILTER  settings\n");
	printf("************************************\n");

	for (x1 = 2; x1 >= 0; x1--)
		if (txCap[x1]) {
			goodTX = x1;
			break;
		}
	if (goodTX < 0) {
		printf("##FATAL ERROR:  Could not determine a supported TX stamping setting\n\n");
		exit(1);
	}
	clrBuf(sock);

	/* testing HWTSTAMP_RX_FILTER settings */
	int gotOne;
	for (x1 = 0; x1 < 15; x1++) {
		gotOne = 0;
		goodIOCTL = -1;
		if (x1 != 2)/* excluding HWTSTAMP_RX_FILTER_SOME */
			goodIOCTL = setIOCTL(sock, iface, goodTX, x1);
		clrBuf(sock);
		/*sender cycling variable amount of packet types
		 * (V1 & V2 & Generic UDP) */
		if (goodIOCTL >= 0) {
			int limit = 15;
			limit = (layer == 2) ? 10 : limit;
			limit = (layer == 5) ? 1  : limit;
			for (x2 = 0; x2 < limit; x2++) {
				res = 0;
				res = select(sock+1, &readfs, 0, &errorfs, &delta);
				delta.tv_sec = 2;
				delta.tv_usec = 0;

				if  (res > 0) {
					res = recvpacket(sock, 0, layer);
					gotOne = 1;
				}
			}
			if (!gotOne)
				printf("##No packets received ...\n");
		} else {
			if ((x1 != 2) && (rxCap[x1]))
				printf("##ERROR : Driver reports this RX FILTER as supported!!\n\n");
		}
	}
	printf("********\n");
	printf("** End of IOCTL Testing.\n\n");
	printf("********\n");

	/* find and set "good" TX/RX settings for setsockopt testing */
	goodTX = -1;
	goodFilter = -1;

	for (x1 = 2; x1 >= 0; x1--)
		if (txCap[x1]) {
			goodTX = x1;
			break;
		}
	if (goodTX < 0) {
		printf("##FATAL ERROR:  Could not determine a supported TX stamping mode\n\n");
		exit(1);
	}

	if (rxCap[HWTSTAMP_FILTER_ALL])
		goodFilter = HWTSTAMP_FILTER_ALL;
	  else {
		printf("HWTSTAMP_FILTER_ALL not supported - looking for an alternative ...\n\n");
		if (rxCap[HWTSTAMP_FILTER_PTP_V2_EVENT]) {
			goodFilter = HWTSTAMP_FILTER_PTP_V2_EVENT;
			fflush(stdin);
			printf("\nStop the current send program on the link partner\n");
			printf("and restart it with the additional command line argument:\n");
			printf("\t-V 2\n");
			printf("PRESS RETURN TO CONTINUE...\n");
			x1 = 0;
			while (!x1)
				x1 = getchar();
			printf("\n");
		}
		  else {
			if (rxCap[HWTSTAMP_FILTER_PTP_V1_L4_SYNC]) {
				if (layer != 2) {
					goodFilter =
						HWTSTAMP_FILTER_PTP_V1_L4_SYNC;
					fflush(stdin);
					printf("\nStop the current send program on the link partner\n");
					printf("and restart it with the additional command line argument:\n");
					printf("\t-V 1\n");
					printf("PRESS RETURN TO CONTINUE...\n");
					x1 = 0;
					while (!x1)
						x1 = getchar();
					printf("\n");
				}
				  else {
					printf("ALL and V2 SYNC packet filters not supported\n");
					printf("and this is a Layer 2 test pass\n\n");
				}
			}
		  }
	  }

	if (goodFilter < 0) {
		printf("Could not find an appropriate filter to test with\n");
		printf("\tRX stamping results are not valid\n");
		goodFilter = HWTSTAMP_FILTER_NONE;
	}

	goodIOCTL = setIOCTL(sock, iface, goodTX, goodFilter);

	clrBuf(sock);

	printf("\n\n************************************\n");
	printf("** Testing setsocketopt ...\n");
	printf("************************************\n");

	int sendToSock;
	signal(SIGINT, catchC);

	for (x1 = 0; x1 <= 2; x1++) {/* TX_HARDWARE and TX_SOFTWARE */
		/* skipping unsupported modes to eliminate redundant tests */
		if ((x1 < 2) && (!sockCap[x1]))
			continue;
		for (x2 = 2; x2 <= 4; x2++) {/* RX_HARDWARE and RX_SOFTWARE */
			/* skipping unsupported modes to eliminate
			 * redundant tests */
			if ((x2 < 4) && (!sockCap[x2]))
				continue;
			/* start off with all reporting modes active
			 * then add stamping modes */
			sendToSock = sockReport;
			/* finishing with no TX stamping mode (x1 == 2) */
			if ((x1 <= 1) && (sockCap[x1]))
				sendToSock |= (1 << x1);
			/* finishing with no RX stamping mode (x2 == 4) */
			if ((x2 <= 3) && (sockCap[x2]))
				sendToSock |= (1 << x2);

			goodSockOpt = setNcheckSock(sock, sendToSock);
			clrBuf(sock);
			delta.tv_sec = 3;
			delta.tv_usec = 0;
			if ((layer == 4) || (layer == 5))
				sendpacket(sock, (struct sockaddr *)&addr,
					sizeof(addr), layer);
			  else
				sendpacket(sock, (struct sockaddr *)&addr2,
					sizeof(addr2), layer);

			res = select(sock+1, &readfs, 0, &errorfs, &delta);

			if (res > 0) {
				res = -1;
				flag = 0;

				for (flag = 0; flag < 1000; flag++) {
					res = recvpacket(sock, MSG_ERRQUEUE,
						layer);
					if (res > 0)
						break;
				}
				if (res <= 0) {
					printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
					printf("No TX packet read from socket!\n");
					printf("\tPacket was not time stamped.\n");
					if ((x1 < 2) && (sockCap[x1]))
						printf("##ERROR: There should of been a TX time stamp!\n\n");
					printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
				}
				res = -1;
				flag = 0;
				while ((res < 0) && (!flag))
					res = recvpacket(sock, 0, layer);
				if (res < 0)
					printf("\n##ERROR reading RX packet from socket!\n");
				if (res == 0)
					printf("\n##WARNING no RX packet read from socket!!\n");
			} else
				printf("\n##ERROR - no packets detected TX or RX!!\n\tCheck Interface\n\n");
		}
	}
	printf("End of setsockopt testing\n\n");
} /* end of function runTest() */

static void queryDriver(int sock, char *iface)
{

	/* BEGIN DRIVER QUERY (ethtool) TO DETERMINE CAPABILITIES */

	struct ethtool_ts_info info;
	struct ifreq ethtool_ifr;

	printf("\n\n************************************\n");
	printf("** Driver reported supported features\n");
	printf("************************************\n\n");

	memset(&ethtool_ifr, 0, sizeof(ethtool_ifr));
	memset(&info, 0, sizeof(info));
	info.cmd = ETHTOOL_GET_TS_INFO;
	strcpy(ethtool_ifr.ifr_name, iface);
	ethtool_ifr.ifr_data = (char *) &info;

	if (ioctl(sock, SIOCETHTOOL, &ethtool_ifr) < 0)
		bail("ioctl SIOCETHTOOL failed");

	printf("Transmit Types:\n");
	if (info.tx_types & (1 << HWTSTAMP_TX_OFF)) {
		printf("\tHWTSTAMP_TX_OFF\n");
		txCap[HWTSTAMP_TX_OFF] = 1;
	}
	if (info.tx_types & (1 << HWTSTAMP_TX_ON)) {
		printf("\tHWTSTAMP_TX_ON\n");
		txCap[HWTSTAMP_TX_ON] = 1;
	}
	if (info.tx_types & (1 << HWTSTAMP_TX_ONESTEP_SYNC)) {
		printf("\tHWTSTAMP_TX_ONESTEP_SYNC\n");
		txCap[HWTSTAMP_TX_ONESTEP_SYNC] = 1;
	}

	printf("\nReceive Filters:\n");

	if (info.rx_filters & (1 << HWTSTAMP_FILTER_NONE)) {
		printf("\tHWTSTAMP_FILTER_NONE\n");
		rxCap[HWTSTAMP_FILTER_NONE] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_ALL)) {
		printf("\tHWTSTAMP_FILTER_ALL\n");
		rxCap[HWTSTAMP_FILTER_ALL] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_SOME)) {
		/* this filter should never show up in the supported filters
		 * result return from ethtool ioctl call.  This is a return
		 * from a SIOCSHWTSTAMP ioctl only value - throw a warning
		 * message if it does show up here */
		printf("\tHWTSTAMP_FILTER_SOME - ERROR\n\n");
		printf("### WARNING ###\n\tThis filter should not be a ");
		printf("supported input\n\tfilter, it is meant to only ");
		printf("be used as a return\n\tvalue from SIOCSHWTSTAMP ");
		printf("ioctl calls!!\n\n");
		rxCap[HWTSTAMP_FILTER_SOME] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V1_L4_EVENT)) {
		printf("\tHWTSTAMP_FILTER_PTP_V1_L4_EVENT\n");
		rxCap[HWTSTAMP_FILTER_PTP_V1_L4_EVENT] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V1_L4_SYNC)) {
		printf("\tHWTSTAMP_FILTER_PTP_V1_L4_SYNC\n");
		rxCap[HWTSTAMP_FILTER_PTP_V1_L4_SYNC] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V1_L4_DELAY_REQ)) {
		printf("\tHWTSTAMP_FILTER_PTP_V1_L4_DELAY_REQ\n");
		rxCap[HWTSTAMP_FILTER_PTP_V1_L4_DELAY_REQ] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L4_EVENT)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L4_EVENT\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L4_EVENT] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L4_SYNC)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L4_SYNC\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L4_SYNC] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L4_DELAY_REQ)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L4_DELAY_REQ\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L4_DELAY_REQ] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L2_EVENT)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L2_EVENT\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L2_EVENT] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L2_SYNC)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L2_SYNC\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L2_SYNC] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_L2_DELAY_REQ)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_L2_DELAY_REQ\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_L2_DELAY_REQ] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_EVENT)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_EVENT\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_EVENT] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_SYNC)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_SYNC\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_SYNC] = 1;
	}
	if (info.rx_filters & (1 << HWTSTAMP_FILTER_PTP_V2_DELAY_REQ)) {
		printf("\tHWTSTAMP_FILTER_PTP_V2_DELAY_REQ\n");
		rxCap[HWTSTAMP_FILTER_PTP_V2_DELAY_REQ] = 1;
	}

	printf("\nTimestamping Reporting:\n");
	if (info.so_timestamping & SOF_TIMESTAMPING_RAW_HARDWARE) {
		printf("\tSOF_TIMESTAMPING_RAW_HARDWARE\n");
		sockReport |= SOF_TIMESTAMPING_RAW_HARDWARE;
	}
	if (info.so_timestamping & SOF_TIMESTAMPING_SYS_HARDWARE) {
		printf("\tSOF_TIMESTAMPING_SYS_HARDWARE\n");
		sockReport |= SOF_TIMESTAMPING_SYS_HARDWARE;
	}
	if (info.so_timestamping & SOF_TIMESTAMPING_SOFTWARE) {
		printf("\tSOF_TIMESTAMPING_SOFTWARE\n");
		sockReport |= SOF_TIMESTAMPING_SOFTWARE;
	}
	printf("\nTimestamp Generation Methods:\n");
	if (info.so_timestamping & SOF_TIMESTAMPING_RX_SOFTWARE) {
		printf("\tSOF_TIMESTAMPING_RX_SOFTWARE\n");
		sockCap[3] = 1;
	}
	if (info.so_timestamping & SOF_TIMESTAMPING_RX_HARDWARE) {
		printf("\tSOF_TIMESTAMPING_RX_HARDWARE\n");
		sockCap[2] = 1;
	}
	if (info.so_timestamping & SOF_TIMESTAMPING_TX_SOFTWARE) {
		printf("\tSOF_TIMESTAMPING_TX_SOFTWARE\n");
		sockCap[1] = 1;
	}
	if (info.so_timestamping & SOF_TIMESTAMPING_TX_HARDWARE) {
		printf("\tSOF_TIMESTAMPING_TX_HARDWARE\n");
		sockCap[0] = 1;
	}
	printf("\n\n");
} /* END OF DRIVER QUERY */


int main(int argc, char **argv)
{
	char *interface;
	int enabled = 1;
	int sock1, sock2, sock3;
	int res, sockSet;
	int genPort = 64321;
	int x1;/* loop variable */
	int ip_multicast_loop = 0;

	struct ifreq device;
	struct ip_mreq imr;

	struct in_addr iaddr;

	flag = 0;
	verb = 0;
	sockReport = 0;

	while (1) {
		static struct option long_options[] = {
			{"verbose",	no_argument,		0,	'v'},
			{"port",	required_argument,	0,	'p'},
			{"help",	no_argument,		0,	'h'},
			{0,		0,			0,	0}
		};

		int c, option_index;

		c = getopt_long(argc, argv, "vp:h?",
			long_options, &option_index);

		if (c == -1)
			break;

		switch (c) {
		case 'v':
			verb = 1;
			break;
		case 'p':
			if (optarg[0] == '-')
				usage(argv[0]);
			genPort = atoi(optarg);
			if (genPort <= 1024)
				bail("Port number must be greater than 1024");
			break;
		case '?':
		case 'h':
			usage(argv[0]);
			break;
		default:
			usage(argv[0]);
		}
	}

	if ((optind >= argc) || (argv[optind][0] == '-')) {
		printf("\nMissing required interface name!!\n");
		usage(argv[0]);
	} else {
		interface = argv[optind];
	}

	/*initialize arrays to keep track of device capabilities*/
	memset(txCap, 0, sizeof(txCap));
	memset(rxCap, 0, sizeof(rxCap));
	memset(sockCap, 0, sizeof(sockCap));

	/* setting output to no buffering, this makes running
		the program through a 'tee' more friendly */
	setvbuf(stdout, NULL, _IONBF, BUFSIZ);

	fflush(stdin);
	printf("Start the 'send_L4' program on the link partner\n");
	printf("PRESS RETURN TO CONTINUE...\n");
	x1 = 0;
	while (!x1)
		x1 = getchar();
	printf("\n");

/* Begin setup of socket for L4 communication */
	sock1 = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock1 < 0)
		bail("socket L4");

	memset(&device, 0, sizeof(device));
	strncpy(device.ifr_name, interface, sizeof(device.ifr_name));
	if (ioctl(sock1, SIOCGIFADDR, &device) < 0)
		bail("getting interface IP address");

	/* bind to PTP port* */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(319);/* PTP event port */
	if (bind(sock1,
		 (struct sockaddr *)&addr,
		 sizeof(struct sockaddr_in)) < 0)
		bail("bind");

	/* set multicast group for outgoing packets */
	inet_aton("224.0.1.130", &iaddr);/* alternate PTP domain 1 */
	addr.sin_addr = iaddr;
	imr.imr_multiaddr.s_addr = iaddr.s_addr;
	imr.imr_interface.s_addr =
		((struct sockaddr_in *)&device.ifr_addr)->sin_addr.s_addr;
	if (setsockopt(sock1, IPPROTO_IP, IP_MULTICAST_IF,
		       &imr.imr_interface.s_addr, sizeof(struct in_addr)) < 0)
		bail("set multicast");

	/* join multicast group, loop our own packet */
	if (setsockopt(sock1, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       &imr, sizeof(struct ip_mreq)) < 0)
		bail("join multicast group");

	if (setsockopt(sock1, IPPROTO_IP, IP_MULTICAST_LOOP,
		       &ip_multicast_loop, sizeof(ip_multicast_loop)) < 0)
		bail("loop multicast");

	/* request IP_PKTINFO for debugging purposes */
	if (setsockopt(sock1, SOL_IP, IP_PKTINFO,
		       &enabled, sizeof(enabled)) < 0)
		printf("%s: %s\n", "setsockopt IP_PKTINFO", strerror(errno));

/* END OF SOCKET SETUP FOR L4 COMMUNICATION */

	queryDriver(sock1, interface);

	/* turning on all socket timestamping options and adding in all
	 * reporting options*/
	sockSet = 0;
	for (x1 = 0; x1 < 4; x1++) {
		if (sockCap[x1])
			sockSet |= (1 << x1);
	}
	sockSet += sockReport;

	res = setNcheckSock(sock1, sockSet);

	printf("\n\n\nEnd of initial setup...\nStarting testing pass\n");

	printf("\nRunning tests with a Layer 4 socket:\n\n");
	runTests(sock1, interface, 4);

/* BEGIN SETUP FOR L2 COMMUNICATION */

	const unsigned char ether_broadcast_addr[] = {
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff
	};

	int index;

	fflush(stdin);
	printf("Start the 'send_L2' program on the link partner\n");
	printf("PRESS RETURN TO CONTINUE...");
	x1 = 0;
	while (!x1)
		x1 = getchar();
	printf("\n");

	/* 0x88f7 is PTP communication */
	sock2 = socket(PF_PACKET, SOCK_DGRAM, htons(0x88f7));
	if (sock2 < 0)
		bail("socket L2");

	memset(&device, 0, sizeof(device));
	strncpy(device.ifr_name, interface, sizeof(device.ifr_name));
	if (ioctl(sock2, SIOCGIFINDEX, &device) < 0)
		bail("ioctl SIOCGIFINDEX failed");

	index = device.ifr_ifindex;

	memset(&addr2, 0, sizeof(addr2));

	addr2.sll_family = AF_PACKET;
	addr2.sll_ifindex = index;
	addr2.sll_halen = ETHER_ADDR_LEN;
	addr2.sll_protocol = htons(0x88f7);
	memcpy(addr2.sll_addr, ether_broadcast_addr, ETHER_ADDR_LEN);

	if (bind(sock2, (struct sockaddr *) &addr2, sizeof(addr2)) < 0)
		bail("socket bind failed");

	setNcheckSock(sock2, sockSet);

/* END SETUP FOR L2 COMMUNICATION */
	printf("\n\nRunning tests with a Layer 2 socket\n\n");

	runTests(sock2, interface, 2);

	/* IXGBE does not support (according to hardware team) the stamping
	 * of all packets.  Generic UDP test not applicable when
	 * HWTSTAMP_FILTER_ALL is not a supported filter */
	if (!(rxCap[HWTSTAMP_FILTER_ALL])) {
		printf("Driver for %s does not support HWTSTAMP_FILTER_ALL.\n",
			interface);
		printf("\tSkipping Generic UDP time stamp tests.\n");
		exit(0);
		}

	fflush(stdin);
	printf("Start the 'send_Gen' program on the link partner\n");
	printf("PRESS RETURN TO CONTINUE...\n");
	x1 = 0;
	while (!x1)
		x1 = getchar();
	printf("\n");

/* BEGIN SETUP FOR LAYER 4 GENERIC PACKET COMMUNICATION */
	sock3 = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock3 < 0)
		bail("socket L4");

	memset(&device, 0, sizeof(device));
	strncpy(device.ifr_name, interface, sizeof(device.ifr_name));
	if (ioctl(sock3, SIOCGIFADDR, &device) < 0)
		bail("getting interface IP address");

	/* bind to Generic port */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(genPort);
	if (bind(sock3,
		 (struct sockaddr *)&addr,
		 sizeof(struct sockaddr_in)) < 0)
		bail("bind");

	/* set multicast group for outgoing packets */
	inet_aton("224.0.0.1", &iaddr); /* all hosts on network */
	addr.sin_addr = iaddr;
	imr.imr_multiaddr.s_addr = iaddr.s_addr;
	imr.imr_interface.s_addr =
		((struct sockaddr_in *)&device.ifr_addr)->sin_addr.s_addr;
	if (setsockopt(sock3, IPPROTO_IP, IP_MULTICAST_IF,
		       &imr.imr_interface.s_addr, sizeof(struct in_addr)) < 0)
		bail("set multicast");

	/* join multicast group, loop our own packet */
	if (setsockopt(sock3, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       &imr, sizeof(struct ip_mreq)) < 0)
		bail("join multicast group");

	if (setsockopt(sock3, IPPROTO_IP, IP_MULTICAST_LOOP,
		       &ip_multicast_loop, sizeof(ip_multicast_loop)) < 0)
		bail("loop multicast");

	/* request IP_PKTINFO for debugging purposes */
	if (setsockopt(sock3, SOL_IP, IP_PKTINFO,
		       &enabled, sizeof(enabled)) < 0)
		printf("%s: %s\n", "setsockopt IP_PKTINFO", strerror(errno));

	setNcheckSock(sock3, sockSet);
/* END SETUP FOR LAYER 4 GENERIC PACKET COMMUNICATION */
	printf("\n\nRunning tests on Layer 4 generic packet communication\n\n");

	/*for (x1 = 0; x1 < sizeof(sync); x1++)
		sync[x1] = GEN_PACKET[x1]; */

	/* layer = 5 is being used for a layer 4 non-PTP socket */
	runTests(sock3, interface, 5);

	return 0;
}
